
use tokio::io;
use tokio::net::TcpListener;
use tokio::net::TcpStream;
use std::sync::{Arc, Mutex};
use signal_hook::{consts::SIGUSR1, iterator::Signals};
use std::time::Duration;
use tokio::time::{sleep, timeout};

struct UpstreamState {
    is_up: bool,
    address: String,
}

struct JunctionState {
    upstreams: Vec<UpstreamState>,
    preference: usize,
}

impl JunctionState {
    fn current(&self) -> String {
        for offset in 0..self.upstreams.len() {
            let idx = (self.preference + offset) % self.upstreams.len();
            if self.upstreams[idx].is_up {
                return self.upstreams[idx].address.clone();
            }
        }
        self.upstreams[self.preference].address.clone()  // all upstreams are down
    }

    fn mark(&mut self, idx: usize, is_up: bool) -> bool {
        if self.upstreams[idx].is_up != is_up {
            self.upstreams[idx].is_up = is_up;
            return true;
        }
        false
    }

    fn advance(&mut self) -> String {
        self.preference = (self.preference + 1) % self.upstreams.len();
        self.upstreams[self.preference].address.clone()
    }
}


type SharedState = Arc<Mutex<JunctionState>>;

pub async fn launch(bind: String, upstreams: Vec<String>) {
    let server = TcpListener::bind(bind).await.unwrap();

    let n_upstreams = upstreams.len();
    let upstreams = upstreams.iter().map(|addr| UpstreamState {
        address: addr.clone(),
        is_up: false,
    }).collect();

    let state: SharedState = Arc::new(Mutex::new(JunctionState {
        upstreams,
        preference: 0,
    }));


    // Launch Interrupt signal catcher
    {
        let state = state.clone();
        tokio::spawn(async move {
            let mut signals = Signals::new([SIGUSR1]).unwrap();
            for _ in signals.forever() {
                let mut state = state.lock().unwrap();
                let addr = state.advance();
                println!("USR1 signal received. Preferred upstream: {:?}", addr);
            }
        });
    }

    // Launch upstream probes
    for idx in 0..n_upstreams {
        let state = state.clone();
        tokio::spawn(async move {
            probe_upstream(idx, state).await
        });
    }

    // Listen for incoming connections
    loop {
        let (socket, _) = server.accept().await.unwrap();
        let state = state.clone();
        tokio::spawn(async move {
            handle_client(socket, state).await;
        });
    }
}

async fn handle_client(client: TcpStream, state: SharedState) {
    let address = state.lock().unwrap().current();
    let server = TcpStream::connect(address.as_str()).await.unwrap();

    let (mut eread, mut ewrite) = client.into_split();
    let (mut oread, mut owrite) = server.into_split();

    tokio::spawn(async move { io::copy(&mut eread, &mut owrite).await });
    tokio::spawn(async move { io::copy(&mut oread, &mut ewrite).await });
}

async fn probe_upstream(upstream_idx: usize, state: SharedState) {
    loop {
        let address = state.lock().unwrap().upstreams[upstream_idx].address.clone();
        let upstream = timeout(
            Duration::from_millis(1000),
            TcpStream::connect(address.as_str())
        ).await;
        let is_up = match upstream  {
            Ok(Ok(_)) => true,
            _ => false,
        };
        if state.lock().unwrap().mark(upstream_idx, is_up) {
            if is_up {
                println!("Upstream {:?} is now up.", address)
            } else {
                println!("Upstream {:?} is now down.", address)
            }
        }
        sleep(Duration::from_millis(1000)).await;
    }
}

#[cfg(test)]
mod test;
